from rura.pipeline.processes.temp import TempProcess
from rura.tracker import Tracker
from sklearn.metrics import classification_report
import matplotlib.pyplot as plt
import numpy as np
import mlflow


class SuicideProcess(TempProcess):
    def extra_metrics(self, model, x, y, data_type):
        preds = model.predict(x, data_type=data_type)

        self.__save_pr_curve(model, y, preds, data_type)
        self.__save_probs(model, preds, data_type)
        self.__class_reports(model, y, preds, data_type)

    def __class_reports(self, model, y_true, y_pred, data_type):
        y_pred = np.round(y_pred).flatten()

        report = classification_report(y_true, y_pred, output_dict=True)
        for m in ['precision', 'recall', 'f1-score']:
            for c in ['0', '1', 'macro avg', 'weighted avg']:
                if c in report:
                    mlflow.log_metric(data_type + '_' + m + '_c' + c, report[c][m])
                else:
                    mlflow.log_metric(data_type + '_' + m + '_c' + c, np.nan)

        with open(self.get_output_path(model, file='class_reports.txt'), 'a') as w:
            w.write('\n')
            w.write(data_type + '\n')
            w.write(classification_report(y_true, y_pred))
            w.write('\n\n')

        mlflow.log_artifact(self.get_output_path(model, file='class_reports.txt'))

    def __save_probs(self, model, y_pred, data_type):
        y_pred = y_pred.flatten()

        plt.hist(y_pred, color='steelblue')
        plt.title(data_type + ' Prediction Probabilities')
        plt.savefig(self.get_output_path(model, file=data_type + '_probs_hist.png'))
        plt.clf()
        mlflow.log_artifact(self.get_output_path(model, file=data_type + '_probs_hist.png'))

    def __save_pr_curve(self, model, y_true, y_pred, data_type):
        y_pred = y_pred.flatten()

        from sklearn.metrics import precision_recall_curve, average_precision_score, roc_auc_score
        from inspect import signature

        precision, recall, _ = precision_recall_curve(y_true, y_pred)
        average_precision = average_precision_score(y_true, y_pred)
        roc_auc = roc_auc_score(y_true, y_pred)

        if Tracker.IS_LOGGING:
            mlflow.log_metric(data_type + '_auc', roc_auc)
            mlflow.log_metric(data_type + '_auc_pr', average_precision)

        # In matplotlib < 1.5, plt.fill_between does not have a 'step' argument
        step_kwargs = ({'step': 'post'}
                       if 'step' in signature(plt.fill_between).parameters
                       else {})
        plt.step(recall, precision, color='b', alpha=0.2,
                 where='post')
        plt.fill_between(recall, precision, alpha=0.2, color='b', **step_kwargs)

        plt.xlabel('Recall')
        plt.ylabel('Precision')
        plt.ylim([0.0, 1.05])
        plt.xlim([0.0, 1.0])
        plt.title('Precision-Recall Curve: AP={0:0.2f}; AUC={0:0.2f}'.format(average_precision, roc_auc))
        plt.savefig(self.get_output_path(model, file=data_type + '_pr_curve.png'))
        plt.clf()
        mlflow.log_artifact(self.get_output_path(model, file=data_type + '_pr_curve.png'))



import importlib


class DatasetMaker:
    def __init__(self, datasets):
        self.datasets = datasets

    # TODO - Check dependencies
    def get_dataset(self, data):
        parts = data.split('/')
        section = parts[0]
        title = parts[1]

        config = self.datasets[section][title]
        if config is None:
            parameters = {}
        else:
            parameters = config.get('parameters', {})
        parameters['path'] = [section, title]
        dataset = importlib.import_module('.' + section + '.make_dataset', package='src.data').MakeDataset(**parameters)

        if not dataset.is_complete():
            print('Making dataset ' + section + '/' + title + '...')
            dataset.make_files(config.get('split', None))
            print('Done!')

        return dataset



from rura.utils.paths import PROJECT
from dotenv import load_dotenv
import mlflow
from mlflow.exceptions import RestException
import os


class Tracker:
    URI = None
    USERNAME = None
    PASSWORD = None
    loaded = False
    IS_LOGGING = False
    client = None

    @classmethod
    def __load(cls):
        if cls.loaded:
            return

        path = os.path.join(PROJECT, '.env')
        if not os.path.exists(path):
            raise ValueError('.env file does not exist: ' + str(path))

        load_dotenv(dotenv_path=path)
        cls.URI = os.getenv('TRACKER_URI')
        cls.USERNAME = os.getenv('TRACKER_URI')
        cls.PASSWORD = os.getenv('TRACKER_URI')

    @classmethod
    def setup(cls):
        if not cls.IS_LOGGING:
            return

        if not cls.loaded:
            cls.__load()
            mlflow.set_tracking_uri(cls.URI)
            cls.client = mlflow.tracking.MlflowClient()
            cls.loaded = True

    @classmethod
    def set_experiment(cls, name):
        # if not cls.IS_LOGGING:
        #     return

        mlflow.set_experiment(name)

    @classmethod
    def does_run_exist(cls, run_id):
        if not cls.IS_LOGGING:
            return

        if run_id is None:
            return False

        try:
            cls.client.get_run(run_id)
        except RestException as e:
            if str(e).startswith('RESOURCE_DOES_NOT_EXIST'):
                return False
            else:
                raise e

        return True

    @classmethod
    def get_artifact(cls, run_id, artifact):
        if not cls.IS_LOGGING:
            return

        if not cls.loaded:
            cls.setup()

        cls.client.download_artifacts(run_id, artifact, 'tmp')

    @classmethod
    def resume_run(cls, run_id):
        return mlflow.start_run(run_id=run_id)

    @classmethod
    def start_run(cls, name, nested_id=None):
        # if not cls.IS_LOGGING:
        #     return None

        is_nested = nested_id is not None
        return mlflow.start_run(run_name=name, nested=is_nested)

    @classmethod
    def end_run(cls):
        if not cls.IS_LOGGING:
            return None

        mlflow.end_run()

    @classmethod
    def disable_logging(cls):
        cls.IS_LOGGING = False

    @classmethod
    def enable_logging(cls):
        cls.IS_LOGGING = True

    @classmethod
    def is_logging(cls):
        return cls.IS_LOGGING



from rura.pipeline.base import Base
from rura.utils.file import save_file


class Transform(Base):
    def __init__(self, all_files=False, output_type='npy', has_y=True, **kwargs):
        super().__init__(**kwargs)
        self.all_files = all_files
        self.output_type = output_type
        self.has_y = has_y

    def load(self):
        pass

    def transform(self, data, data_type):
        raise NotImplementedError

    def save_data(self, file, file_type, data):
        save_file(self.get_path(file + '.' + file_type), file_type, data)
        self.files[file] = file_type



from rura.pipeline.base import Base
from rura.utils.dataset import split_patients
from rura.utils.file import read_file
import os


class Dataset(Base):
    def __init__(self, version, files, **kwargs):
        super().__init__(**kwargs)
        self.version = version
        self.files = files
        self._get_files()

    def load(self):
        pass

    def _get_files(self):
        pass

    def get_path(self, file=None):
        path = super().get_path(file=file)
        if file is None:
            return path

        if self.version is not None:
            parts = os.path.splitext(path)
            return parts[0] + '.v' + str(self.version) + parts[1]

        return path

    # def get_data(self, name):
    #     if isinstance(name, list):
    #         return tuple([read_file(self.get_path(file), self.files[file]) for file in name])
    #
    #     if name not in self.files:  # TODO - Is this an ok long term solution?
    #         for file in self.files.keys():
    #             if file.startswith(name):
    #                 name = file
    #                 break
    #
    #     return read_file(self.get_path(name), self.files[name])

    def make(self):
        raise NotImplementedError

    def make_files(self, split):
        df = self.make()
        if split is not None:
            if isinstance(split, dict):
                seed = split.get('seed', 3)
            else:
                seed = 3
            df_train, df_val, df_test = split_patients(df, seed=seed)
            df_train = df_train.reset_index()
            df_val = df_val.reset_index()
            df_test = df_test.reset_index()

            print('Train/Val/Test Suicide Patient Counts')
            print(df_train.groupby('PatientID')['Output'].max().sum())
            print(df_val.groupby('PatientID')['Output'].max().sum())
            print(df_test.groupby('PatientID')['Output'].max().sum())

            df_train.to_feather(self.get_path('train'))
            df_val.to_feather(self.get_path('val'))
            df_test.to_feather(self.get_path('test'))
        else:
            raise NotImplementedError



import pandas as pd
import numpy as np


def read_file(path, ext, options={}):
    if ext == 'feather':
        return pd.read_feather(path, **options)
    elif ext == 'csv':
        return pd.read_csv(path, **options)
    elif ext == 'npy':
        try:
            return np.load(path, **options)
        except ValueError:
            return np.load(path, allow_pickle=True, **options)
    elif ext == 'txt':
        with open(path) as f:
            return f.read()
    else:
        raise NotImplementedError


# TODO - Stop computing the final path within here
def save_file(path, ext, data):
    if ext == 'feather':
        data.to_feather(path)
    elif ext == 'csv':
        data.to_csv(path, index=False)
    elif ext == 'npy':
        np.save(path, data)
    elif ext == 'txt':
        with open(path, 'w') as f:
            f.write(data)
    else:
        raise NotImplementedError



from rura.utils.paths import TEMP
import datetime
import os


def get_temp_file_path(ext):
    if not os.path.exists(TEMP):
        os.makedirs(TEMP)

    suffix = datetime.datetime.now().strftime("%y%m%d_%H%M%S")

    return os.path.join(TEMP, 'temp-' + suffix + '.' + ext)



from rura.utils.paths import DATA_PROCESSED, DATA_INTERIM
from rura.utils.file import read_file
import os


class Base:
    def __init__(self, path, folder=DATA_PROCESSED, sources=None):
        self.folder = folder

        self.sources = sources

        self.files = {}
        self.path = path

    def load(self):
        raise NotImplementedError

    def load_files(self):
        for file in os.listdir(self.get_path()):
            parts = os.path.splitext(file)
            self.files[parts[0]] = parts[1].replace('.', '')

    def get_path(self, file=None):
        path = os.path.join(self.folder, *self.path)
        if not os.path.exists(path):
            os.makedirs(path)

        if file is None:
            return path

        if file not in self.files:
            return os.path.join(path, file)

        return os.path.join(path, file + '.' + self.files[file])

    def get_data(self, file, allow_none=False):
        if not isinstance(file, list):
            return read_file(self.get_path(file), self.files[file])

        data = []
        for f in file:
            if f not in self.files:
                if allow_none:
                    data.append(None)
                    continue
                raise FileNotFoundError('Unable to find file: ' + f)

            data.append(read_file(self.get_path(f), self.files[f]))

        return data

    def is_complete(self):
        if self.files is None or len(self.files.keys()) == 0:  # TODO - This is temp for transform and model
            return len(list(os.listdir(self.get_path()))) > 0

        for file in self.files:
            if not os.path.exists(self.get_path(file)):
                return False

        # TODO - Delete old files if they exist

        return True



import os

PROJECT = os.getcwd()
if 'notebooks' in PROJECT:
    PROJECT = PROJECT.split('notebooks')[0]
REPORTS = os.path.join(PROJECT, 'reports')
RESULTS = os.path.join(PROJECT, 'results')
DATA = os.path.join(PROJECT, 'data')
DATA_EXTERNAL = os.path.join(DATA, 'external')
DATA_RAW = os.path.join(DATA, 'raw')
DATA_INTERIM = os.path.join(DATA, 'interim')
DATA_PROCESSED = os.path.join(DATA, 'processed')
TEMP = os.path.join(PROJECT, 'tmp')



from rura.logic.class_def import ClassDef
from rura.logic.dataset import DatasetMaker
from rura.logic.model import ModelExecute
from rura.logic.sources import Sources
from rura.logic.transform import Transformer
from rura.utils.paths import PROJECT, DATA_PROCESSED, RESULTS
from ruamel.yaml import YAML
import warnings
import logging
import shutil
import os

warnings.simplefilter("ignore", category=DeprecationWarning)


class Runner:
    yaml = YAML()
    yaml.indent(mapping=2, sequence=4, offset=2)

    def __init__(self, file, analysis):
        self.config = None
        self.file = file
        logging.basicConfig(format='%(levelname)s: %(message)s')
        self.__load()

        Sources.RUNNER = self
        self.analysis_name = analysis
        self.current_pipeline_name = None
        self.items = {}
        self.dataset = DatasetMaker(self.config['datasets'])
        self.model_execute = ModelExecute(self)

    def __load(self):
        with open(os.path.join(PROJECT, self.file)) as f:
            self.config = self.yaml.load(f)

        # Convert hashes and sources to be strings
        for title, pipeline in self.config['pipelines'].items():
            for i in range(len(pipeline)):
                pipeline[i]['hash'] = str(pipeline[i]['hash'])

                if 'source' in pipeline[i]:
                    if isinstance(pipeline[i]['source'], list):
                        pipeline[i]['source'] = [str(source) for source in pipeline[i]['source']]
                    else:
                        pipeline[i]['source'] = str(pipeline[i]['source'])

                if pipeline[i]['type'] == 'model':
                    for j in range(len(pipeline[i]['runs'])):
                        pipeline[i]['runs'][j]['hash'] = str(pipeline[i]['runs'][j]['hash'])

    def __save(self):
        with open(os.path.join(PROJECT, self.file), 'w') as w:
            self.yaml.dump(self.config, w)

    def clean_dataset(self, title):
        path = os.path.join(DATA_PROCESSED, self.analysis_name, title)
        if os.path.exists(path):
            shutil.rmtree(path)

    def clean_pipeline(self, title):
        path = os.path.join(DATA_PROCESSED, self.analysis_name, title)
        if os.path.exists(path):
            shutil.rmtree(path)

        path = os.path.join(RESULTS, self.analysis_name, title)
        if os.path.exists(path):
            shutil.rmtree(path)

    def execute_pipeline(self, title, until=None):
        self.current_pipeline_name = title
        for item in self.config['pipelines'][title]:
            if until is not None and item['hash'] == str(until):
                print('Hit until point: ' + str(until) + '; stopping.')
                break

            self.execute(item)

        print('Completed pipeline ' + title)

    def evaluate_model(self, pipeline, model_hash, run_hash):
        self.current_pipeline_name = pipeline
        model_item = None
        for item in self.config['pipelines'][pipeline]:
            if item['hash'] == model_hash:
                model_item = item
            self.execute(item, execute=False)

        if not isinstance(run_hash, list):
            run_hash = [run_hash]

        for run in run_hash:
            model = self.items[model_hash][run]
            self.model_execute.evaluate_model(model_item, model, run)
            print('Done!')

    def get_dataset(self, title):
        return self.dataset.get_dataset(title)

    def execute(self, item, execute=True):
        class_def = ClassDef.get_class(item)
        for run in item['runs']:
            parameters = {**item.get('parameters', {}), **run.get('parameters', {})}
            if 'sources' in run or 'source' in run:
                parameters['sources'] = Sources.get_sources(run, self.items)
            else:
                parameters['sources'] = Sources.get_sources(item, self.items)
            parameters['path'] = [self.analysis_name, self.current_pipeline_name, item['hash'], run['hash']]

            if item['hash'] not in self.items:
                self.items[item['hash']] = {}

            if item['type'] == 'transform':
                item_instance = class_def(**parameters)
                self.items[item['hash']][run['hash']] = item_instance
                if execute:
                    Transformer.transform(item_instance)
            elif item['type'] == 'model':
                model = self.model_execute.run_model(item, run, class_def, parameters, execute=execute)
                self.items[item['hash']][run['hash']] = model
            else:
                logging.error('Undefined type "' + item['type'] + '".')



from rura.pipeline.process import Process
from rura.utils.file import save_file


class TrainConvertProcess(Process):
    def train(self, model):
        # Get the train and val data
        x_train, x_val, x_test = model.sources[0].get_data(['train_x', 'val_x', 'test_x'])

        model.set_parameters_data(x_train, None, x_val, None, x_test, None)

        # Build the model
        model.build()

        # Officially log parameters
        model.log_parameters()

        # Fit the model (send validation as an optional parameter since some models dont use it)
        model.fit(x_train, None, x_val=x_val, y_val=None)

        model.fit_metrics(x_train, None, x_val=x_val, y_val=None)

        model.log_model()
        return model

    def run(self, model):
        trained_model = self.train(model)
        self.convert(trained_model)

    def convert(self, model):
        x_train, x_val, x_test = model.sources[0].get_data(['train_x', 'val_x', 'test_x'])
        for data_type, data in zip(['train', 'val', 'test'], [x_train, x_val, x_test]):
            # TODO - Change this to also be a model.save_data() similar to transform ability
            save_file(model.get_path(data_type + '_x.npy'), 'npy', model.predict(data))



from sklearn.model_selection import train_test_split


def split_patients(df, seed):
    patient_ids = df['PatientID'].unique()
    train_ids, test_ids = train_test_split(patient_ids, test_size=0.2, random_state=seed)
    train_ids, val_ids = train_test_split(train_ids, test_size=0.2, random_state=seed)

    return df[df['PatientID'].isin(train_ids)], df[df['PatientID'].isin(val_ids)], df[
        df['PatientID'].isin(test_ids)]



from rura.pipeline.base import Base


class Model(Base):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.model = None

    def set_parameters_data(self, x_train, y_train, x_val, y_val, x_test, y_test):
        pass

    def log_parameters(self):
        raise NotImplementedError

    def build(self):
        raise NotImplementedError

    def fit(self, x_train, y_train, x_val, y_val):
        raise NotImplementedError

    def fit_metrics(self, x_train, y_train, x_val=None, y_val=None):
        raise NotImplementedError

    def log_model(self):
        raise NotImplementedError

    def predict(self, x):
        raise NotImplementedError



import ipykernel  # Importing to make keras logging work
from rura.pipeline import Model
import mlflow.keras
from keras.models import load_model


class KerasModel(Model):
    def log_parameters(self):
        mlflow.keras.autolog()

    def log_model(self):
        # TODO - Add this file to the self.files
        self.model.save(self.get_path('model.h5'))

    def load(self):
        self.model = load_model(self.get_path('model.h5'))



from rura.utils.file import read_file, save_file
from rura.utils.paths import DATA_EXTERNAL
import os


# TODO - Make the folder allowed to be an array and pull it apart
class ExternalData:
    @staticmethod
    def get_ext(file):
        return os.path.splitext(file)[1][1:]

    @classmethod
    def get_data(cls, folder, file, options={}):
        ext = cls.get_ext(file)
        return read_file(os.path.join(DATA_EXTERNAL, folder, file), ext, options=options)

    @classmethod
    def save_data(cls, folder, file, data):
        ext = cls.get_ext(file)
        save_file(os.path.join(DATA_EXTERNAL, folder), ext, data)



class Sources:
    RUNNER = None

    @classmethod
    def __get_list(cls, item):
        if 'source' in item:
            sources = item['source']
        elif 'sources' in item:
            sources = item['sources']
        else:
            raise ValueError('Sources not defined for this item.')

        if not isinstance(sources, list):
            sources = [sources]

        return [cls.__get_info(source) for source in sources]

    @classmethod
    def __get_info(cls, source):
        if source.startswith('data:'):
            return {
                'type': 'dataset',
                'title': source.replace('data:', ''),
            }

        pipeline = cls.RUNNER.config['pipelines'][cls.RUNNER.current_pipeline_name]
        model_sub = None
        if ':' in source:
            parts = source.split(':')
            source = parts[0]
            model_sub = parts[1]

        for item in pipeline:
            if item['hash'] == source:
                return {
                    'type': item['type'],
                    'title': source,
                    'config': item,
                    'run_id': model_sub,
                }

        raise ValueError('Unable to find the source.')

    @classmethod
    def get_sources(cls, item, dictionary):
        sources = cls.__get_list(item)

        data = []
        for source in sources:
            if source['title'] in dictionary:
                if source['run_id'] is None:
                    raise ValueError('No run_id specified for source: ' + source['config']['hash'])
                data.append(dictionary[source['title']][source['run_id']])
            elif source['type'] == 'dataset':
                data.append(cls.RUNNER.get_dataset(source['title']))
            else:
                raise ValueError('Unable to find current source: ' + str(source))

        return data



class Transformer:
    @staticmethod
    def transform(transformer):
        if transformer.is_complete():
            # transformer.load_files()
            # transformer.load()
            print('Skipping transformer ' + '/'.join(transformer.path) + '; already exists.')
            return

        for source in transformer.sources:
            source.load_files()
            source.load()

        dataset = transformer.sources[0]

        print('Running transformer ' + '/'.join(transformer.path) + '...')
        if transformer.all_files:
            file_names = ['train_x', 'train_y', 'val_x', 'val_y', 'test_x', 'test_y']
            files = transformer.transform(dataset.get_data(file_names), None)
            for file, data in zip(file_names, files):
                transformer.save_data(file, 'npy', data)
        else:
            for data_type in ['train', 'val', 'test']:
                file = data_type
                if data_type not in dataset.files:  # TODO - Figure out how to do this better in the future
                    for f in dataset.files:
                        if f.startswith(data_type):
                            file = f
                            break
                result = transformer.transform(dataset.get_data(file), data_type)
                if transformer.has_y:
                    x, y = result
                    transformer.save_data(data_type + '_y', transformer.output_type, y)
                else:
                    x = result

                transformer.save_data(data_type + '_x', transformer.output_type, x)

        print('Done!')



from setuptools import setup
from setuptools import find_packages

setup(name='rura',
      version='0.2.0',
      packages=find_packages(),
      license='MIT',
      description='Pipelines for machine learning',
      long_description=open('README.md').read(),
      long_description_content_type='text/markdown',
      author='Filip Dabek',
      url='https://github.com/fdabek1/rura',
      download_url='https://github.com/fdabek1/rura/archive/0.2.0.tar.gz',
      keywords=['machine learning', 'pipeline', 'etl'],
      install_requires=[
          'mlflow',
          'numpy>=1.9.1',
          'scipy>=0.14',
          'six>=1.9.0',
          'pyyaml',
          'h5py',
          'python-dotenv',
      ],
      extras_require={
          'tests': ['pytest',
                    'pytest-pep8',
                    'pytest-xdist',
                    'flaky',
                    'pytest-cov',
                    'pandas',
                    'requests',
                    'markdown'],
      },
      classifiers=[  # Optional
          # How mature is this project? Common values are
          #   3 - Alpha
          #   4 - Beta
          #   5 - Production/Stable
          'Development Status :: 3 - Alpha',

          # Indicate who your project is intended for
          'Intended Audience :: Developers',
          'Topic :: Software Development :: Build Tools',

          # Pick your license as you wish
          'License :: OSI Approved :: MIT License',

          # Specify the Python versions you support here. In particular, ensure
          # that you indicate whether you support Python 2, Python 3 or both.
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.8',
      ])



from rura.pipeline.process import Process


class VecNNProcess(Process):
    # TODO - Remove this class by making the regular training process automatically make any y values be None
    def run(self, model):
        # Get the train and val data
        x_train, y_train, x_val, y_val, x_test, y_test = model.sources[0].get_data(
            ['train_x', 'train_y', 'val_x', 'val_y', 'test_x', 'test_y'])

        # Set parameters dynamically
        model.set_parameters_data(x_train, y_train, x_val, y_val, x_test, y_test)

        # Build the model
        model.build()

        # Officially log parameters
        model.log_parameters()

        # Fit the model (send validation as an optional parameter since some models dont use it)
        model.fit(x_train, y_train, x_val=x_val, y_val=y_val)

        model.fit_metrics(x_train, x_train, x_val=x_val, y_val=y_val)

        model.log_model()



from rura.pipeline.processes.convert import TrainConvertProcess
from rura.pipeline.processes.suicide import SuicideProcess
from rura.pipeline.processes.temp import TempProcess

from rura.pipeline.processes.vec_nn import VecNNProcess
from rura.tracker import Tracker


class ModelExecute:
    def __init__(self, runner):
        self.runner = runner

    def get_process(self, item):
        process_name = item.get('process', None)
        if process_name == 'TrainConvert':
            return TrainConvertProcess()
        elif process_name == 'VecNN':
            return VecNNProcess()
        elif process_name == 'Suicide':
            return SuicideProcess()

        return TempProcess()

    def run_model(self, item, run, model_class, parameters, execute=True):
        Tracker.setup()
        exp_name = self.runner.analysis_name + ':' + self.runner.current_pipeline_name + ':'
        exp_name += str(item.get('base', 'None')) + ':' + str(item['func']) + ':' + item['hash']
        Tracker.set_experiment(exp_name)

        if execute:
            print('Running model ' + item['hash'] + '...')
        process = self.get_process(item)

        model = model_class(**parameters)

        if not execute:
            return model

        if model.is_complete():
            print('Skipping run ' + run['hash'] + '; already exists.')
            # model.load_files()
            # model.load()
            return model

        # TODO - Save the run_id to the yaml file
        # if Tracker.is_logging():
        run_id = Tracker.start_run(run['hash']).info.run_id

        for source in model.sources:
            source.load_files()
            source.load()

        print('Processing run ' + run['hash'] + '...')
        process.run_id = run['hash']
        process.run(model)

        Tracker.end_run()
        return model

    def evaluate_model(self, item, model, run_hash):
        from mlflow.tracking.client import MlflowClient
        from mlflow.entities import ViewType

        exp_name = self.runner.analysis_name + ':' + self.runner.current_pipeline_name + ':'
        exp_name += str(item.get('base', 'None')) + ':' + str(item['func']) + ':' + item['hash']

        client = MlflowClient()
        experiments = [exp for exp in client.list_experiments() if exp.name == exp_name]
        if len(experiments) == 0 or len(experiments) > 1:
            raise ValueError('Unable to find the experiment.')
        experiment = experiments[0]

        run = client.search_runs(
            experiment_ids=experiment.experiment_id,
            filter_string='tags."mlflow.runName" = ' + "'" + run_hash + "'",
            run_view_type=ViewType.ACTIVE_ONLY,
            max_results=1,
        )[0]

        run_id = run.info.run_id
        Tracker.resume_run(run_id)

        process = self.get_process(item)

        for source in model.sources:
            source.load_files()
            source.load()
        model.load()

        print('Evaluating run ' + run_hash + '...')
        process.run_id = run_hash
        process.evaluate(model)

        Tracker.end_run()



from rura.pipeline.process import Process
from rura.utils.paths import RESULTS
import os


class TempProcess(Process):
    def run(self, model):
        # Get the train and val data
        x_train, y_train, x_val, y_val, x_test, y_test = model.sources[0].get_data(
            ['train_x', 'train_y', 'val_x', 'val_y', 'test_x', 'test_y'], allow_none=True)

        # TODO - Rename to compute_parameters_data
        model.set_parameters_data(x_train, y_train, x_val, y_val, x_test, y_test)

        # Build the model
        model.build()

        # Officially log parameters
        # TODO - Auto log stuff from the init function
        model.log_parameters()

        # Fit the model (send validation as an optional parameter since some models dont use it)
        model.fit(x_train, y_train, x_val=x_val, y_val=y_val)

        model.fit_metrics(x_train, y_train, x_val=x_val, y_val=y_val)

        model.log_model()

        if not os.path.exists(self.get_output_path(model)):
            os.makedirs(self.get_output_path(model))

        self.extra_metrics(model, x_train, y_train, 'train')
        self.extra_metrics(model, x_val, y_val, 'val')

    @staticmethod
    def get_output_path(model, file=''):
        return os.path.join(RESULTS, *model.path, file)

    def extra_metrics(self, model, x, y, data_type):
        pass

    def evaluate(self, model):
        for filename in os.listdir(os.path.join(RESULTS, *model.path)):
            if filename.startswith('test_'):
                print('Skipping evaluation due to already evaluated.')
                return

        # Get the test data
        x_test, y_test = model.sources[0].get_data(['test_x', 'test_y'], allow_none=True)
        self.extra_metrics(model, x_test, y_test, 'test')



import importlib
import inspect


class ClassDef:
    @staticmethod
    def get_package(item):
        if item['type'] == 'dataset':
            return 'src.data'
        elif item['type'] == 'transform':
            return 'src.features'
        elif item['type'] == 'model':
            return 'src.models'

        raise ImportError('Unknown type.')

    @classmethod
    def get_class(cls, item):
        package = cls.get_package(item)
        func = item['func']
        base = item.get('base', None)
        if base is None:
            builder = importlib.import_module('._base.' + func, package=package)
        else:
            builder = importlib.import_module('.' + base + '.' + func, package=package)

        for m in inspect.getmembers(builder, inspect.isclass):
            class_def = m[1]
            if class_def.__module__.startswith(package):
                return class_def

        raise ImportError('Unable to find the class definition.')
